public class GrafMA {
  private Object[] varfuri; // tabloul referintelor la varfuri
  private byte[][] arce; // matricea de adiacenta
  int numarVarfuri;

  public GrafMA(int numarMaximVarfuri) {
    varfuri=new Object[numarMaximVarfuri];
    arce=new byte[numarMaximVarfuri][numarMaximVarfuri];
    numarVarfuri=0;
  }

  public boolean existaVarf(Object varf) {
    if(varf==null) return false;
    for(int k=0; k<numarVarfuri; k++)
      if(varf.equals(varfuri[k])) return true;
    return false;
  }

  public int indiceVarf(Object varf) {
    if(varf==null) return -1;
    for(int k=0; k<numarVarfuri; k++)
      if(varf.equals(varfuri[k])) return k;
    return -1;
  }

  public boolean puneVarf(Object varf) throws Exception {
    if(varf==null) return false;
    if(numarVarfuri==varfuri.length)
      throw new Exception("Depasire numar maxim varfuri");
    for(int i=0; i<numarVarfuri; i++)
      if(varf==varfuri[i]) return false;
    varfuri[numarVarfuri]=varf;
    numarVarfuri++;
    return true;
  }

  public boolean puneArc(Object baza, Object varf) 
                             throws Exception {
    int i=-1, j=-1;
    i=indiceVarf(baza);
    if(i<0) throw new Exception("Varf inexistent: "+baza);
    j=indiceVarf(varf);
    if(j<0) throw new Exception("Varf inexistent: "+varf);
    if(arce[i][j]==1) return false;
    arce[i][j]=1;
    return true;
  }


  public boolean puneArc(int indiceBaza, int indiceVarf) {
    if(indiceBaza<0 || indiceBaza>=numarVarfuri || 
      indiceVarf<0 || indiceVarf>=numarVarfuri) return false;
    int i=indiceBaza, j=indiceVarf;
    if(arce[i][j]==1) return false;
    arce[i][j]=1;
    return true;
  }

  public boolean eliminaVarf(Object varf) {
    if(varf==null) return false;
    int k=indiceVarf(varf);
    if(k<0) return false;
    return eliminaVarf(k);
  }

    public boolean eliminaVarf(int indice) {
    if(indice>=numarVarfuri) return false;
    int k=indice, nrVarfuriRedus=numarVarfuri-1;
    if(k<0) return false;
    /* Se deplaseaza la stanga toate coloanele matricei arce cu j>k
       si se pune 0 pe locul ramas liber in dreapta;
    for(int i=0; i<numarVarfuri; i++)
      for(int j=k+1; j<numarVarfuri; j++) {
        arce[i][j-1]=arce[i][j];
        arce[i][nrVarfuriRedus]=0;
      }
    /* Se deplaseaza in sus toate liniile matricei arce cu i>k
       si se pune 0 pe locul ramas liber jos
    */
    for(int j=0; j<numarVarfuri; j++)
      for(int i=k+1; i<numarVarfuri; i++) { 
        arce[i-1][j]=arce[i][j];
        arce[nrVarfuriRedus][j]=0;
      }
    // Se deplaseaza la stanga toate varfurile cu indice>k
    for(int i=k+1; i<numarVarfuri; i++) varfuri[i-1]=varfuri[i];
    varfuri[nrVarfuriRedus]=null;
    numarVarfuri=nrVarfuriRedus;
    return true;
  }

  public boolean eliminaArc(Object baza, Object varf) {
    if(baza==null || varf==null) return false;
    return eliminaArc(indiceVarf(baza), indiceVarf(varf));
  }
    
  public boolean eliminaArc(int indiceBaza, int indiceVarf) {
    int i=indiceBaza, j=indiceVarf;
    if(i<0 || i>=numarVarfuri || j<0 || j>=numarVarfuri)
       return false;
    if(arce[i][j]==0) return false;
    arce[i][j]=0;
    return true;
  }

  // Alte metode ale clasei GrafMA

  public String toString() {
    StringBuffer buffer=new StringBuffer();
    buffer.append("{ Varfuri: [");
    for(int i=0; i<numarVarfuri; i++) 
      buffer.append("("+varfuri[i]+")");
    buffer.append("] Matrice adiacente: [");
    for(int i=0; i<numarVarfuri; i++) {
      buffer.append("[");
      for(int j=0; j<numarVarfuri-1; j++) 
        buffer.append(arce[i][j]+", ");
      if(numarVarfuri>0) buffer.append(arce[i][numarVarfuri-1]);
      if(i<numarVarfuri-1) buffer.append("], ");
      else buffer.append("]");
    }
    buffer.append("]}");
    return buffer.toString();    
  }
}
