/* Cautarea prin metoda programarii dinamice a caii de lungime minima intre 
   doua varfuri ale unui graf orientat 
*/

import java.util.*;

class DynamicProg1 {

  /* Clasa imbricata a arcelor grafului */
  public static class ArcGraf {
    String origine, varf;
    double lungime;

    ArcGraf(String origine, String varf, double lungime) {
      this.origine=origine;
      this.varf=varf;
      this.lungime=lungime;
    }
  }

  /* Clasa imbricata a grafului orientat, reprezentat printr-o lista de arce */
  public static class Graf {
    List listaArce;

    public Graf(ArrayList listaArce) {
      this.listaArce=new ArrayList(listaArce);
    }

    public void puneArc(String origine, String varf, double lungime) {
      listaArce.add(new ArcGraf(origine, varf, lungime));
    }

    public List listaArce() {
      return listaArce;
    }
  }    

  /* Clasa imbricata a varfurilor arborelui de acoperire */
  static class VarfArbore {
    String eticheta;
    VarfArbore tata;
    double distanta;
    int nivel;

    VarfArbore(String varfCurent, VarfArbore tata, double distanta, int nivel) {
      eticheta=varfCurent;
      this.tata=tata;
      this.distanta=distanta;
      this.nivel=nivel;
    }

    public boolean equals(Object obj) {
      VarfArbore varf=(VarfArbore)obj;
      return eticheta.equals(varf.eticheta);
    }

    public String toString() {
      String etata;
      if(tata==null) etata="null";
      else etata=tata.eticheta;
      return "("+eticheta+", tata: "+etata+" dist:"+distanta+
        " nivel: "+nivel+")";
    }
  }

  /* Clasa imbricata pentru solutia problemei. Solutia contine atat arborele de
     acoperire de distanta minima, cat si calea de la varful initial la cel final
  */
  static class Solutie {
    List arbore, cale;

    public String toString() {
      return "arbore: "+arbore+"\ncale: "+cale;
    }
  }

  /* Metoda statica de obtinere a solutiei prin trasare inapoi */
  public static Solutie trasareInapoi(Graf graf, String varfInitial, String varfFinal) {
    ArrayList arbore=new ArrayList();
    ArrayList cale=new ArrayList();
    LinkedList listaArce=new LinkedList(graf.listaArce());
    VarfArbore varfCurent, varfTinta;
    ArcGraf arcCurent;
    /* Se pune in arbore varful final ca radacina */
    boolean epuizatNivel=false;
    if(varfFinal.equals(varfInitial)) epuizatNivel=true;
    arbore.add(new VarfArbore(varfFinal, null, 0, 0));
    /* incepe ciclul pe niveluri */
    int indiceNivel=0, inceputNivel=0, sfarsitNivel=0;
    boolean epuizatArce=false;
    while(!epuizatArce) {
      /* Incepe ciclul pentru nivelul curent */
      indiceNivel++;
      epuizatArce=true;
      for(int i=inceputNivel; i<=sfarsitNivel; i++) {
        varfTinta=(VarfArbore)arbore.get(i);
        Iterator iter=listaArce.iterator();
        /* Se prelucreaza toate arcele care se termina intr-un nod al nivelului curent */
        while(iter.hasNext()) {
          arcCurent=(ArcGraf)iter.next();
          if(arcCurent.varf.equals(varfTinta.eticheta)) {
            /* Se cauta daca varful de origine al arcului curent deja exista in arbore */
            epuizatArce=false;
            boolean exista=false;
            for(int j=0; !exista && j<arbore.size(); j++) {
              VarfArbore varfExistent=((VarfArbore)arbore.get(j));
              double distantaNoua=arcCurent.lungime+varfTinta.distanta;
              if(arcCurent.origine.equals(varfExistent.eticheta)) {
                exista=true;
                if(distantaNoua<varfExistent.distanta) {
                  varfExistent.distanta=distantaNoua;
                  varfExistent.tata=varfTinta;
                  varfExistent.nivel=indiceNivel;
                } // sfarsit if
              }  // sfarsit if
            } // sfarsit for
            if(!exista) 
              arbore.add(new VarfArbore(arcCurent.origine, varfTinta,
                 arcCurent.lungime+varfTinta.distanta, indiceNivel));   
            iter.remove();
          } 
        }
      } // sfarsit for nivel curent
      inceputNivel=sfarsitNivel+1;
      sfarsitNivel=arbore.size()-1;
    } // sfarsit while niveluri

    /* Din arbore se constituie calea */
    /* Se cauta in arbore varful initial */
    VarfArbore sursa=null;
    int index=arbore.indexOf(new VarfArbore(varfInitial, null, 0, 0));
    if(index>=0) {
      /* se pune in cale varful initial */
      sursa=(VarfArbore)arbore.get(index);
      cale.add(sursa);
      /* Se pun in cale toate celelalte varfuri, pana la cel final */
      varfCurent=sursa;
      while(varfCurent.tata!=null) {
        varfCurent=varfCurent.tata;
        cale.add(varfCurent);
      }
    }
    Solutie solutie=new Solutie();
    solutie.arbore=arbore;
    solutie.cale=cale;
    return solutie;
  } 

  /* Metoda statica de obtinere a solutiei prin trasare inainta */
  public static Solutie trasareInainte(Graf graf, String varfInitial, String varfFinal) {
    ArrayList arbore=new ArrayList();
    ArrayList cale=new ArrayList();
    LinkedList listaArce=new LinkedList(graf.listaArce());
    VarfArbore varfCurent, varfSursa;
    ArcGraf arcCurent;
    /* Se pune in arbore varful initial ca radacina */
    boolean epuizatNivel=false;
    if(varfFinal.equals(varfInitial)) epuizatNivel=true;
    arbore.add(new VarfArbore(varfInitial, null, 0, 0));
    /* incepe ciclul pe niveluri */
    int indiceNivel=0, inceputNivel=0, sfarsitNivel=0;
    boolean epuizatArce=false;
    while(!epuizatArce) {
      /* Incepe ciclul pentru nivelul curent */
      indiceNivel++;
      epuizatArce=true;
      for(int i=inceputNivel; i<=sfarsitNivel; i++) {
        varfSursa=(VarfArbore)arbore.get(i);
        Iterator iter=listaArce.iterator();
        /* Se prelucreaza toate arcele care pornesc dintr-un nod al nivelului curent */
        while(iter.hasNext()) {
          arcCurent=(ArcGraf)iter.next();
          if(arcCurent.origine.equals(varfSursa.eticheta)) {
            /* Se cauta daca varful de destinatie al arcului curent deja exista in arbore */
            epuizatArce=false;
            boolean exista=false;
            for(int j=0; !exista && j<arbore.size(); j++) {
              VarfArbore varfExistent=((VarfArbore)arbore.get(j));
              double distantaNoua=arcCurent.lungime+varfSursa.distanta;
              if(arcCurent.varf.equals(varfExistent.eticheta)) {
                exista=true;
                if(distantaNoua<varfExistent.distanta) {
                  varfExistent.distanta=distantaNoua;
                  varfExistent.tata=varfSursa;
                  varfExistent.nivel=indiceNivel;
                } // sfarsit if
              }  // sfarsit if
            } // sfarsit for
            if(!exista) 
              arbore.add(new VarfArbore(arcCurent.varf, varfSursa,
                 arcCurent.lungime+varfSursa.distanta, indiceNivel));   
            iter.remove();
          }
        }
      } // sfarsit for nivel curent
      inceputNivel=sfarsitNivel+1;
      sfarsitNivel=arbore.size()-1;
    } // sfarsit while niveluri

    /* Din arbore se constituie calea */
    /* Se cauta in arbore varful final */
    VarfArbore sursa=null;
    int index=arbore.indexOf(new VarfArbore(varfFinal, null, 0, 0));
    if(index>=0) {
      /* se pune in cale varful final */
      sursa=(VarfArbore)arbore.get(index);
      cale.add(sursa);
      /* Se pun in cale toate celelalte varfuri, pana la cel final */
      varfCurent=sursa;
      while(varfCurent.tata!=null) {
        varfCurent=varfCurent.tata;
        cale.add(varfCurent);
      }
    }
    Solutie solutie=new Solutie();
    solutie.arbore=arbore;
    solutie.cale=cale;
    return solutie;
  } 

}



